// Compact Disc Detect & Execute
// 
// Copyright(C) 2002-03 Eric Lathrop <eric@ericlathrop.com>
//

// regular headers
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <stdlib.h>
#include <linux/cdrom.h>
#include <fstab.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <getopt.h>

// libxml headers
#include <libxml/parser.h>
#include <libxml/tree.h>

// my headers
#include "config.h"
#include "cdde.h"
#include "string.h"

// signal handler to find out when out child has exited
void sigchld(int signum)
{
	int status;
	wait(&status);

	// if there are still children waiting 
	// re-install the signal handler
	numchildren--;
	if (numchildren > 0)
	{
		// re-install the signal handler
		signal(SIGCHLD, sigchld);
	}
}

// signal handler to find out when to re-read config file
void sighup(int signum)
{
	struct stat fileinfo;

	if (verbose) printf("Caught SIGHUP, re-reading config file.\n");
	
	// free old stuff
	while (list_size(drives) > 0)
	{
		removedrive(0);
	}
	
	// make sure the config file is there
	if (stat(filename, &fileinfo) == -1)
	{
		switch (errno)
		{
			case ENOENT:
				fprintf(stderr, "Error: Config file %s does not exist!\n", filename);
				if (saveconfig(filename))
				{
					fprintf(stderr, "Error: couldn't create template file.\n");
				} else {
					printf("I created a template configuration file located at %s\n", filename);
					printf("I suggest you edit it to make sure the values are useful before you run cdde again.\n");
				}
				break;
			case EACCES:
				fprintf(stderr, "Error: I can't read from config file %s!\n", filename);
				break;
		}
		
		exit (1);
	}
	
	// read the config file
	if (loadconfig(filename))
	{
		fprintf(stderr, "Error: Can't read config file %s\nUnable to continue.\n", filename);
		keeprunning = 0;
	}

	// make sure there's a drive to check
	if (list_size(drives) == 0)
	{
		fprintf(stderr, "Warning: No drives to poll!\n");
	}

	// re-install the signal handler
	signal(SIGHUP, sighup);
}

// catch a ^c so we can clean up
void sigint(int signum)
{
	if (verbose) printf("Caught SIGINT, exiting gracefully.\n");
	keeprunning = 0;
}

// fork and exec a command
// wait for it to exit before returning
void execute(const char * command)
{
	char * fixed_str;
	char ** args;
	int num_args;
	int curr_arg;
	int temp_len;
	int found_start;
	int pos;
	
	// exit if nothing to execute;
	if (command == "") return;

#if 1
	system(command);
#else
	numchildren++;
	signal(SIGCHLD, sigchld);

	if (fork() == 0)
	{
		// im the child
		// i get to execute the command

		// trim whitespace from the command
		fixed_str = (char *) strdup(command);
		fixed_str = trim_destruct(&fixed_str);

		// find the number of arguments
		num_args = 0;
		found_start = 0;
		for (pos=0; pos<strlen(fixed_str); pos++)
		{
			if ((!found_start) && (fixed_str[pos] != ' '))
			{
				found_start = 1;
				num_args++;
			}

			if ((found_start) && (fixed_str[pos] == ' '))
			{
				found_start = 0;
			}
		}

		// create the argument array
		args = malloc((num_args + 1) * sizeof(char *));
		found_start = 0;
		curr_arg = 0;
		temp_len = strlen(fixed_str);
		for (pos=0; pos<temp_len; pos++)
		{
			if ((!found_start) && (fixed_str[pos] != ' '))
			{
				found_start = 1;
				args[curr_arg] = &fixed_str[pos];
				curr_arg++;
			}

			if ((found_start) && (fixed_str[pos] == ' '))
			{
				found_start = 0;
				fixed_str[pos] = '\0';
			}
		}
		args[curr_arg] = NULL;

		// call execvp
		execvp(args[0], args);
		
		// should never get here
		perror("Error: Executing command failed");
		exit(2);
	}
#endif
	// i'm the parent
	// i already have a signal handler to tell me when a child dies
	// so I can just get on with my business
}

// executes all commands for a related disc type
void executeall(drive * d, int cmdtype)
{
	int i;
	char * command;
	
	// get fstab info on where the device's mount point is
	struct fstab * cdfs = getfsspec(d->filename);
	char * mountpoint;
	if (cdfs == NULL)
	{
		fprintf(stderr, "Error: couldn't get %s entry for %s\nAborting executing commands...\n", _PATH_FSTAB, d->filename);
		return;
	} else {
		mountpoint = substr(cdfs->fs_file, 0, strlen(cdfs->fs_file));
	}
	
	for (i=list_size(d->commands[cmdtype])-1; i>=0; i--)
	{
		command = replace(list_nth(d->commands[cmdtype], i), "%dev%", d->filename);
		command = replace_destruct(&command, "%mnt%", mountpoint);
		execute(command);
		free(command);
	}
}

// determine if the device has a directory in its top level
//
// device is the filename of the special device file
// directory is the name of the directory to look for in the discs top directory
//
// returns nonzero if the file exists
// otherwise returns 0
int disc_has_dir(const char * device, const char * directory)
{
	int ret = 0; // return value
	int fd = 0; // file descriptor for drive
	unsigned short bs; // the discs block size
	unsigned short ts; // the path table size
	unsigned int tl; // the path table location (in blocks)
	int i;
	unsigned int len_di = 0; // length of the directory name in current path table entry
	unsigned int parent = 0; // the number of the parent directory's path table entry
	char * dirname = NULL; // filename for the current path table entry
	int pos = 0; // our position into the path table
	int curr_record = 1; // the path table record we're on
	char * fixed_directory = toupper(directory); // the uppercase version of the "directory" parameter
	
	// open the drive
	fd = open(device, O_RDONLY | O_NONBLOCK);
	if (fd == -1)
	{
		fprintf(stderr, "Error: Failed to open device %s", device);
		return 0;
	}

	// read the block size
	lseek(fd, 0x8080, SEEK_CUR);
	read(fd, &bs, 2);

	// read in size of path table
	lseek(fd, 2, SEEK_CUR);
	read(fd, &ts, 2);
	
	// read in which block path table is in
	lseek(fd, 6, SEEK_CUR);
	read(fd, &tl, 4);
	
	// seek to the path table
	lseek(fd, ((int)(bs) * tl), SEEK_SET);
	
	// loop through the path table entries
	while (pos < ts)
	{
		// get the length of the filename of the current entry
		read(fd, &len_di, 1);

		// get the record number of this entry's parent
		// i'm pretty sure that the 1st entry is always the top directory
		lseek(fd, 5, SEEK_CUR);		
		read(fd, &parent, 2);
		
		// allocate and zero a string for the filename
		dirname = (char *) malloc(len_di+1);
		for (i=0; i<len_di+1; i++) dirname[i] = '\0';
		
		// read the name
		read(fd, dirname, len_di);
		dirname = toupper_destruct(&dirname);
		
		// if we found a folder that has the root as a parent, and the directory name matches
		// then return success
		if ((parent == 1) && (strcmp(dirname, fixed_directory) == 0))
		{
			ret = 1;
			free(dirname);
			break;
		}
		
		free(dirname);
	
		// all path table entries are padded to be even, so if this is an odd-length table, seek a byte to fix it
		if (len_di%2 == 1)
		{
			lseek(fd, 1, SEEK_CUR);
			pos++;
		}
		
		// update our position
		pos += 8+len_di;
		curr_record++;
	}
		
	close(fd);
	free(fixed_directory);
	return ret;
}

// see if a drive has a disc in it
// if so, determine type of disc
// and execute proper command
//
// returns 0 on ok and nonzero on error
int checkdrive(drive * d)
{
	int fd;
	int status;
	struct cdrom_tochdr th;
	char * errmsg = NULL;
	
	// open the device
	fd = open(d->filename, O_RDONLY | O_NONBLOCK);
	if (fd < 0)
	{
		//printf("Error: couldn't open %s: ", filename);
		errmsg = replace("Error: couldn't open %s", "%s", d->filename);
		perror(errmsg);
		free(errmsg);
		return -1;
	}

	// read the drive status info
	status = ioctl(fd, CDROM_DRIVE_STATUS, CDSL_CURRENT);
	
	switch (status)
	{
		// if there's a ok disc in there
		case CDS_DISC_OK:
			if (d->dontexecute)
			{
				// release the device
				close(fd);
				break;
			}

			d->dontexecute = 1;

			// see if we can read the disc's table of contents (TOC).
			status = ioctl(fd, CDROMREADTOCHDR, &th);
			if (status != 0)
			{
				// release the device
				close(fd);

				printf("Can't read disc TOC. The disc is either a blank, or has a broken TOC.\n");
				executeall(d, DATA_BLANK);
				break;
			}

			// read disc status info
			status = ioctl(fd, CDROM_DISC_STATUS, CDSL_CURRENT);

			// release the device
			close(fd);

			switch (status)
			{
				case CDS_AUDIO:
					// found a audio cd
					printf("Detected audio cd!\n");
					executeall(d, DATA_AUDIO);
					break;
				case CDS_DATA_1:
				case CDS_DATA_2:
					// found a data cd
					printf("Detected data cd!\n");

					if (disc_has_dir(d->filename, "video_ts") != 0)
					{
						printf("I think it's a DVD...\n");
						executeall(d, DATA_DVD);
					} else if (disc_has_dir(d->filename, "vcd") != 0) {
						printf("I think it's a VCD...\n");
						executeall(d, DATA_VCD);
					} else if (disc_has_dir(d->filename, "svcd") != 0) {
						printf("I think it's a SVCD...\n");
						executeall(d, DATA_SVCD);
					} else {
						executeall(d, DATA_DATA);
					}
					break;
				case CDS_MIXED:
					// found a mixed cd
					printf("Detected mixed audio/data cd!\n");
					executeall(d, DATA_MIXED);
					break;
				default:
					printf("Could not determine disc type: Doing nothing!\n");
					break;
			}
			break;
		case CDS_NO_INFO:
			// drive doesnt support querying, so this program will never work on that drive.
			fprintf(stderr, "Warning: %s does not support status queries.\n", filename);
		default:
			// release the device
			d->dontexecute = 0;
			close(fd);
	}
	return 0;
}

// read a configuration file
// returns nonzero on error
int loadconfig(char * filename)
{
	xmlDocPtr doc = NULL;
	xmlNodePtr cur = NULL;
	char * temp = NULL;
	
	// parse the document
	doc = xmlParseFile(filename);
	if (doc == NULL)
	{
		fprintf(stderr, "Error: Could not parse document.\n");
		xmlFreeDoc(doc);
		return 1;
	}

	// get the root node
	cur = xmlDocGetRootElement(doc);
	if (cur == NULL)
	{
		fprintf(stderr, "Error: Could not parse document: unable to find root element.\n");
		xmlFreeDoc(doc);
		return 1;
	}

	// make sure the root is a "cdde"
	if (xmlStrcmp(cur->name, (const xmlChar *) "cdde"))
	{
		fprintf(stderr, "Error: Could not parse document: root element is not \"cdde\".\n");
		xmlFreeDoc(doc);
		return 1;
	}

	// find the delay
	temp = xmlGetProp(cur, (const xmlChar *) "delay");
	if (temp == NULL)
	{
		fprintf(stderr, "Warning: Did not find delay in config file, using default.\n");
		delay = DEFAULT_SLEEP_N;
	} else {
		errno = 0;
		delay = strtol(temp, NULL, 10);
		if (errno != 0)
		{
			fprintf(stderr, "Warning: Delay had under/overflow, using default.\n");
			delay = DEFAULT_SLEEP_N;
		}
		if (delay < 0)
		{
			fprintf(stderr, "Warning: Delay can't be negative, using default.\n");
			delay = DEFAULT_SLEEP_N;
		}
		
		xmlFree(temp);
	}
	if (verbose) printf("Using delay of %d microseconds.\n", delay);
	
	// find all of the children
	cur = cur->xmlChildrenNode;
	while (cur != NULL)
	{
		if (!xmlStrcmp(cur->name, (const xmlChar *) "drive"))
		{
			parsedrive(cur);
		}
		cur = cur->next;
	}
	xmlFreeDoc(doc);
	
	return 0;
}

// parse the <drive> nodes
// and their children
void parsedrive(xmlNodePtr cur)
{
	int i;
	drive  * d = malloc(sizeof(drive));
	xmlChar * data;
	
	if (d == NULL)
	{
		fprintf(stderr, "Error: Couldn't allocate memory for drive!.\n");
		exit(1);
	}
	for (i=0; i<NUM_DATA_TYPES; i++)
	{
		d->commands[i] = NULL;
	}
	d->dontexecute = dontrunfirst;
	
	d->filename = xmlGetProp(cur, "path");
	
	// search it's children for commands, and add them
	cur = cur->xmlChildrenNode;
	while (cur != NULL)
	{
		for (i=0; i<NUM_DATA_TYPES; i++)
		{
			if (!xmlStrcmp(cur->name, (const xmlChar *) datatags[i]))
			{
				data = xmlGetProp(cur, "command");
				d->commands[i] = list_push(d->commands[i], data);
				
				break;
			}
		}
		cur = cur->next;
	}
	drives = list_push(drives, d);
}

// write a template configuation file
// returns zero on success
int saveconfig(char * filename)
{
	xmlDocPtr doc;
	xmlNodePtr cur;
	xmlNodePtr root;
	xmlNodePtr subroot;
	int retval = 0;
	
	// create the root
	root = xmlNewNode(NULL, (xmlChar *) "cdde");
	xmlSetProp(root, (xmlChar *) "delay", (xmlChar *) DEFAULT_SLEEP_S);

	// add a drive
	// FIXME: try to guess cdrom device paths by looking at fstab
	subroot = xmlNewNode(NULL, (xmlChar *) "drive");
	xmlSetProp(subroot, (xmlChar *) "path", (xmlChar *) "/dev/cdrom");
	xmlAddChild(root, subroot);

	// add commands
	cur = xmlNewNode(NULL, (xmlChar *) "audio");
	xmlSetProp(cur, (xmlChar *) "command", (xmlChar *) "echo A audio cd was inserted.");
	xmlAddChild(subroot, cur);

	// add commands
	cur = xmlNewNode(NULL, (xmlChar *) "data");
	xmlSetProp(cur, (xmlChar *) "command", (xmlChar *) "echo A data cd was inserted.");
	xmlAddChild(subroot, cur);

	// add commands
	cur = xmlNewNode(NULL, (xmlChar *) "dvd");
	xmlSetProp(cur, (xmlChar *) "command", (xmlChar *) "echo A dvd was inserted.");
	xmlAddChild(subroot, cur);

	// add commands
	cur = xmlNewNode(NULL, (xmlChar *) "vcd");
	xmlSetProp(cur, (xmlChar *) "command", (xmlChar *) "echo A vcd was inserted.");
	xmlAddChild(subroot, cur);

	// add commands
	cur = xmlNewNode(NULL, (xmlChar *) "svcd");
	xmlSetProp(cur, (xmlChar *) "command", (xmlChar *) "echo A svcd was inserted.");
	xmlAddChild(subroot, cur);

	// add commands
	cur = xmlNewNode(NULL, (xmlChar *) "blank");
	xmlSetProp(cur, (xmlChar *) "command", (xmlChar *) "echo A blank cdr was inserted.");
	xmlAddChild(subroot, cur);

	// add commands
	cur = xmlNewNode(NULL, (xmlChar *) "mixed");
	xmlSetProp(cur, (xmlChar *) "command", (xmlChar *) "echo A mixed (audio/data) cd was inserted.");
	xmlAddChild(subroot, cur);

	// create the document
	doc = xmlNewDoc("1.0");
	xmlDocSetRootElement(doc, root);

	// write out the xml document
	if (xmlSaveFormatFile (filename, doc, 1) == -1) retval = -1;

	// free memory
	xmlFreeDoc(doc);
	
	return retval;
}

// parse command line arguments
void parseargs(int argc, char ** argv)
{
	int c;
	int longindex = 0;
	
	struct option long_options[] =
	{
		{ "help", 0, NULL, '?' },
		{ "version", 0, NULL, 'V' },
		{ "verbose", 0, NULL, 'v' },
		{ "config", 1, NULL, 'c' },
		{ "run", 0, NULL, 'r' },
		{ "batch", 0, NULL, 'b' },
		{ 0, 0, 0, 0 }
	};

	while ((c = getopt_long(argc, argv, "c:Vvhrb", long_options, &longindex)) != -1)
	{
		switch (c)
		{
			case 'c':
				filename = substr(optarg, 0, strlen(optarg));
				break;
			case 'V':
				printf("%s\n", VERSION);
				exit(0);
				break;
			case 'v':
				verbose = 1;
				break;
			case 'r':
				dontrunfirst = 0;
				break;
			case 'b':
				dontrunfirst = 0;
				batchmode = 1;
				break;
			case 'h':
			case '?':
			default:
				printf("Usage: %s [-c filename] [-rbvVh]\n", PACKAGE);
				printf("Run commands when discs are inserted into CD/DVD drives.\n\n");

				printf(" -r\t--run\t\tRun commands on disc already in drive when is started.\n");
				printf(" -b\t--batch\t\tBatch mode: check all drives once then exit.\n");
				printf("\t\t\t(implies -r)\n");
				printf(" -c\t--config\tUse an alternate config file. (Default is ~/.cdde.xml)\n");
				printf(" -V\t--version\tPrint out version information.\n");
				printf(" -v\t--verbose\tPrint out lots of information.\n");
				printf(" -h\t--help\t\tPrint this help message.\n");
				exit(0);
		}
	}
}

// remove a drive from the list
void removedrive(unsigned int drivenum)
{
	xmlChar * data;
	drive * d;
	int i;
	
	drives = list_removenth(drives, (void **)&d, drivenum);
	
	for (i=0; i<NUM_DATA_TYPES; i++)
	{
		while (list_size(d->commands[i]) > 0)
		{
			d->commands[i] = list_pop(d->commands[i], (void **)&data);
			xmlFree(data);
		}
	}
	free(d);
}

// the program's starting point
int main(int argc, char ** argv)
{
	struct stat fileinfo;
	int i=0;

	parseargs(argc, argv);
	
	if (filename == NULL)
	{
		// create string representing where the config file is
		filename = (char *) strdup(getenv("HOME"));
		filename = realloc(filename, ((strlen(filename)+1) + strlen("/.cdde.xml")) * sizeof(char));
		filename = (char *) strcat(filename, "/.cdde.xml");
	}
	if (verbose) printf("Using config file: %s\n", filename);

	// make sure the config file is there
	if (stat(filename, &fileinfo) == -1)
	{
		switch (errno)
		{
			case ENOENT:
				fprintf(stderr, "Error: Config file %s does not exist!\n", filename);
				if (saveconfig(filename))
				{
					fprintf(stderr, "Error: couldn't create template file.\n");
				} else {
					printf("I created a template configuration file located at %s\n", filename);
					printf("I suggest you edit it to make sure the values are useful before you run cdde again.\n");
				}
				break;
			case EACCES:
				fprintf(stderr, "Error: I can't read from config file %s!\n", filename);
				break;
		}
		return 0;
	}

	// install signal handlera
	signal(SIGHUP, sighup);
	signal(SIGINT, sigint);

	// read the config file
	if (loadconfig(filename))
	{
		fprintf(stderr, "Error reading config file %s\nUnable to continue.\n", filename);
		keeprunning = 0;
	}

	// make sure there's a drive to check
	if (list_size(drives) == 0)
	{
		fprintf(stderr, "Warning: No drives to poll!\n");
	}
	
	// pause for "delay" microseconds and check the drive forever
	while (keeprunning)
	{
		for (i=0; i<list_size(drives); i++)
		{
			if (checkdrive((drive *)list_nth(drives, i)) != 0)
			{
				removedrive(i);
				i--;
				if (list_size(drives) == 0)
				{
					fprintf(stderr, "Warning: all drives were found invalid. Nothing left to do!\n");
					keeprunning = 0;
				}
			}
		}
		if (batchmode) break;
		usleep(delay);
	}

	// cleanup time!
	while (list_size(drives) > 0)
	{
		removedrive(0);
	}
	free(filename);
	
	return 0;
}
